from typing import Any, Awaitable, Callable, List, MutableMapping, Tuple

Headers = List[Tuple[bytes, bytes]]
Scope = MutableMapping[str, Any]
Message = MutableMapping[str, Any]

Receive = Callable[[], Awaitable[Message]]
Send = Callable[[Message], Awaitable[None]]
ASGIApp = Callable[[Scope, Receive, Send], Awaitable[None]]



from qactuar.config import Config
from qactuar.models import ASGIApp
from qactuar.server import QactuarServer


def make_server(
    host: str = None, port: int = None, app: ASGIApp = None, conf: Config = None,
) -> QactuarServer:
    qactuar_server = QactuarServer(host, port, app, conf)
    return qactuar_server


def run(
    host: str = None, port: int = None, app: ASGIApp = None, conf: Config = None,
) -> None:
    qactuar_server = make_server(host, port, app, conf)
    qactuar_server.serve_forever()



from dataclasses import dataclass, field
from datetime import datetime
from email.utils import formatdate
from time import mktime

from qactuar.__version__ import VERSION
from qactuar.models import Headers
from qactuar.util import BytesList


@dataclass
class Response:
    status: bytes = b"200"
    headers: Headers = field(default_factory=list)
    body: BytesList = field(default_factory=BytesList)

    def to_http(self) -> bytes:
        headers = [
            (b"Date", formatdate(mktime(datetime.now().timetuple())).encode("utf-8")),
            (b"Server", b"Qactuar " + VERSION.encode("utf-8")),
        ] + self.headers
        response = BytesList()
        response.writelines(b"HTTP/1.1 ", self.status, b"\r\n")
        for header in headers:
            key, value = header
            response.writelines(key, b": ", value, b"\r\n")
        response.write(b"\r\n")
        response.writelines(self.body.readlines())
        return response.read()



import json
import os
from dataclasses import dataclass, field
from logging import getLogger
from pathlib import Path
from typing import Dict


@dataclass
class Config:
    HOST: str = "localhost"
    PORT: int = 8000
    LOG_LEVEL: str = "DEBUG"
    CHECK_PROCESS_INTERVAL: int = 1
    SELECT_SLEEP_TIME: float = 0.025
    RECV_TIMEOUT: float = 0.001
    RECV_BYTES: int = 65536
    MAX_PROCESSES: int = 500
    APPS: Dict[str, str] = field(default_factory=dict)


def config_init() -> Config:
    logger = getLogger("Qactuar")
    path_str = os.environ.get("QACTUAR_CONFIG_DIR")
    config_path = Path(path_str or "")
    if not config_path.is_file():
        logger.warning(
            f"Config file path is not valid or not set, loading default values. To use "
            f"a config file set up an environment variable called 'QACTUAR_CONFIG_DIR' "
            f"and set the value to a JSON config file path."
        )
        return Config()
    return Config(**json.loads(config_path.open().read()))



import sys

import qactuar

if __name__ == "__main__":
    if len(sys.argv) < 2:
        qactuar_server = qactuar.make_server()
        qactuar_server.serve_forever()
    app_path = sys.argv[1]
    module, app = app_path.split(":")
    module = __import__(module)
    app = getattr(module, app)
    qactuar.run("127.0.0.1", 8000, app)



import argparse
import threading
from time import time

import requests

RESPONSES = []
ELAPSEDES = []
THREADS = []
LOREM_IPSUM = """Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nulla eu commodo justo. Interdum et malesuada fames ac ante ipsum primis in faucibus. Aliquam erat volutpat. Aliquam risus eros, mattis et volutpat eget, gravida vitae justo. Nunc risus leo, tincidunt in egestas vitae, dapibus sed enim. Interdum et malesuada fames ac ante ipsum primis in faucibus. Sed posuere libero vitae tortor mollis interdum sed quis risus. Phasellus sed turpis at augue luctus efficitur vitae vitae urna. Duis in lacus vitae tortor efficitur elementum a sed mauris. Morbi pharetra feugiat faucibus. Quisque risus dui, venenatis at placerat ut, dictum consequat lacus. Donec egestas hendrerit quam quis eleifend.

Nulla dictum blandit ligula, a interdum ipsum iaculis ut. Ut porta, orci sed ultricies tempor, felis metus sodales felis, at convallis urna nisi in arcu. Pellentesque quis condimentum dolor, in volutpat mauris. Donec sollicitudin tempus felis, non aliquam ante fringilla vel. Nulla suscipit rhoncus erat ac vehicula. Nullam lacinia sollicitudin suscipit. Nullam faucibus placerat sem ut pretium. Nunc at aliquet nulla. Duis vitae porta orci. Etiam consequat ipsum eu massa laoreet, et tincidunt lectus ultrices. Vivamus id iaculis libero, ac porttitor nunc. Integer faucibus nibh velit, ac bibendum ipsum lobortis id. Morbi ut consectetur sem.

Aliquam erat volutpat. Cras in vulputate nisl. Curabitur quis purus neque. Vivamus a faucibus arcu. Sed sit amet dignissim orci. Nulla at condimentum turpis, sed tristique tellus. Vestibulum dictum imperdiet nibh ut lacinia. Vestibulum scelerisque odio est, sed congue enim hendrerit quis. Quisque venenatis orci non venenatis pellentesque. Integer consectetur tristique lacus, at sagittis erat tincidunt eu. Nunc varius, lectus in consequat tincidunt, purus erat consectetur sem, ut aliquet tellus neque vitae purus. Sed ultricies massa id eleifend faucibus. Duis laoreet ac justo auctor porta. Nunc nec magna in sapien congue scelerisque. In velit dui, placerat ut quam nec, semper sagittis dui. Nam vel sem ultricies, iaculis justo mattis, ultricies felis.

Sed tempus vehicula rutrum. Morbi porta magna ut purus placerat vulputate. Phasellus efficitur dolor eget elit cursus congue. Quisque sagittis eget diam quis finibus. Praesent dignissim velit ac quam sagittis, ut rutrum lorem venenatis. Vivamus aliquet ex vitae ex ornare auctor. In hac habitasse platea dictumst.

Mauris nec est quis massa sollicitudin tempus eu et nisi. Nunc luctus lacus facilisis quam laoreet lacinia. Morbi sagittis ligula quis ligula ornare tempus vel vel neque. Vivamus porta scelerisque metus. Interdum et malesuada fames ac ante ipsum primis in faucibus. Morbi vitae blandit erat. Nullam nec iaculis dolor."""


def connect(num):
    response = requests.post(
        "http://localhost:8000/what", data={"thing": "stuff", "num": num}
    )
    elapsed = response.elapsed
    ELAPSEDES.append(elapsed.microseconds)
    response = response.text
    # RESPONSES.append(response)
    print(f"{response} {elapsed}")


def start(max_clients, max_conns):
    start_time = time()
    for client_num in range(max_clients):
        for connection_num in range(max_conns):
            t = threading.Thread(target=connect, args=(client_num,))
            t.daemon = True
            THREADS.append(t)
    for thread in THREADS:
        thread.start()
    for thread in THREADS:
        thread.join()
    for response in RESPONSES:
        print(response)
    end_time = time()
    total_time = end_time - start_time
    print(f"total time: {total_time}, avg: {(sum(ELAPSEDES)/len(ELAPSEDES))/1000}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Test client for Qactuar.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--max-conns",
        type=int,
        default=1,
        help="Maximum number of connections per client.",
    )
    parser.add_argument(
        "--max-clients", type=int, default=1, help="Maximum number of clients."
    )
    args = parser.parse_args()
    start(args.max_clients, args.max_conns)



import urllib.parse
from typing import List, Tuple

from qactuar.header import Header
from qactuar.util import BytesList


class Request:
    def __init__(self, request: bytes = None):
        self._raw_request: bytes = request or b""
        self._command: str = ""
        self._request_version: str = ""
        self._path: str = ""
        self._raw_path: bytes = b""
        self._query_string: bytes = b""
        self._headers = []
        self._parsed_headers: Header = Header()
        self._body: bytes = b""
        self.headers_complete = False
        if request:
            self.parse()

    def parse(self):
        lines = self._raw_request.split(b"\r\n")

        command, self._path, request_version = lines.pop(0).split(b" ")
        self._command = command.decode("utf-8")
        self._request_version = request_version.decode("utf-8")

        path_parts = self._path.split(b"?")
        self._path = path_parts[0].decode("utf-8")
        self._raw_path = path_parts[0]
        if len(path_parts) > 1:
            query_string = path_parts[1]
        else:
            query_string = b""
        self._query_string = query_string

        line_num = 0
        for line in lines:
            if not line:
                self.headers_complete = True
                break
            line_num += 1
            key, value = line.split(b": ", 1)
            self._headers.append((key.lower(), value))
        self._parsed_headers = Header(self._headers)

        body = BytesList()
        body.writelines(lines[line_num:])
        self._body = body.read()

    @property
    def headers(self) -> Header:
        return self._parsed_headers

    @property
    def raw_headers(self) -> List[Tuple[bytes, bytes]]:
        return self._headers

    @property
    def request_version(self) -> str:
        return self._request_version

    @property
    def request_version_num(self) -> str:
        return self._request_version.replace("HTTP/", "")

    @property
    def command(self) -> str:
        return self._command

    @property
    def path(self) -> str:
        return urllib.parse.unquote(self._path)

    @path.setter
    def path(self, path: str) -> None:
        self._path = path

    @property
    def raw_path(self) -> bytes:
        return self._raw_path

    @property
    def query_string(self) -> bytes:
        return self._query_string

    @property
    def body(self) -> bytes:
        return self._body

    @property
    def raw_request(self) -> bytes:
        return self._raw_request

    @raw_request.setter
    def raw_request(self, request: bytes) -> None:
        self._raw_request = request
        try:
            self.parse()
        except ValueError:
            self._reset_values()

    def _reset_values(self) -> None:
        self._command: str = ""
        self._request_version: str = ""
        self._path: str = ""
        self._raw_path: bytes = b""
        self._query_string: bytes = b""
        self._headers = []
        self._parsed_headers: Header = Header()
        self._body: bytes = b""
        self.headers_complete = False



from qactuar.__version__ import VERSION
from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    requirements = fh.readlines()

setup(
    name="Qactuar",
    version=VERSION,
    description="ASGI compliant web server",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Anthony Post",
    author_email="postanthony3000@gmail.com",
    license="MIT License",
    url="https://github.com/Ayehavgunne/Qactuar/",
    packages=find_packages(),
    install_requires=requirements,
    extras_require={
        "dev": ["mypy", "black", "isort", "requests", "tonberry", "pytest"]
    },
    python_requires=">=3.7",
)



from tonberry import File, create_app, expose
from tonberry.content_types import TextHTML, TextPlain


class Root:
    @expose.get
    async def index(self) -> TextPlain:
        return "Hello, how are you?"

    @expose.get
    async def what(self, thing, num) -> TextPlain:
        return f"Hello {thing} {num}"

    @expose.post
    async def what(self, thing, num) -> TextPlain:
        return f"Go away {thing} {num}"

    @expose.get
    async def li(self) -> TextHTML:
        return File("../tests/lorem_ipsum.html")


app = create_app(Root)



from typing import Dict, Optional, Union

from qactuar.models import Headers


class Header:
    def __init__(self, header: Headers = None):
        if header is None:
            header = []
        self._header = header
        self._header_attrs: Dict[str, str] = {}
        self.decode()

    def __contains__(self, item: str) -> bool:
        if item in self._header_attrs:
            return True
        return False

    def __getitem__(self, item: str) -> Optional[str]:
        if item in self._header_attrs:
            return self._header_attrs[item]
        return None

    def __setitem__(self, key: str, value: Union[str, int]) -> None:
        self._header_attrs[key] = str(value)

    def __delitem__(self, key: str) -> None:
        del self._header_attrs[key]

    def decode(self) -> None:
        for key, value in self._header:
            self._header_attrs[key.decode("utf-8")] = value.decode("utf-8")



from typing import List, Union


class BytesList:
    def __init__(self):
        self._bytes_list: List[bytes] = []

    def write(self, new_bytes: bytes) -> None:
        self._bytes_list.append(new_bytes)

    def writelines(self, *bytes_list: Union[bytes, List[bytes]]) -> None:
        if len(bytes_list) == 1:
            if isinstance(bytes_list[0], (list, tuple)):
                bytes_list = bytes_list[0]
        self._bytes_list.extend(bytes_list)

    def read(self) -> bytes:
        return b"".join(self._bytes_list)

    def readlines(self) -> List[bytes]:
        return self._bytes_list

    def clear(self) -> None:
        self._bytes_list = []

    def __contains__(self, item):
        return item in self.read()



import asyncio
import errno
import logging
import multiprocessing
import select
import socket
import sys
from importlib import import_module
from logging import Logger
from time import time
from typing import Dict, Optional

from qactuar.config import Config, config_init
from qactuar.exceptions import HTTPError
from qactuar.models import ASGIApp, Message, Receive, Scope, Send
from qactuar.request import Request
from qactuar.response import Response
from qactuar.util import BytesList

ASGI_VERSION = {"version": "2.0", "spec_version": "2.0"}
LIFESPAN_SCOPE = {"type": "lifespan", "asgi": ASGI_VERSION}


class QactuarServer(object):
    address_family = socket.AF_INET
    socket_type = socket.SOCK_STREAM
    socket_level = socket.SOL_SOCKET
    socket_opt_name = socket.SO_REUSEADDR
    request_queue_size = 65536

    def __init__(
        self,
        host: str = None,
        port: int = None,
        app: ASGIApp = None,
        config: Config = None,
    ):
        self.config: Config = config or config_init()

        self.logger: Logger = logging.getLogger("Qactuar")
        self.logger.setLevel(getattr(logging, self.config.LOG_LEVEL))
        self.logger.addHandler(logging.StreamHandler())

        self.host: str = host or self.config.HOST
        self.port: int = port or self.config.PORT

        self.listen_socket: socket.socket = socket.socket(
            self.address_family, self.socket_type
        )
        self.listen_socket.setsockopt(self.socket_level, self.socket_opt_name, 1)
        self.listen_socket.bind((self.host, self.port))
        self.listen_socket.listen(self.request_queue_size)

        self.server_name: str = socket.getfqdn(self.host)
        self.server_port: int = self.port

        self.client_connection: Optional[socket.socket] = None
        self.scheme: str = "http"
        self.raw_request_data: bytes = b""
        self.request_data: Request = Request()
        self.response: Response = Response()
        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        self.processes: Dict[int, multiprocessing.Process] = {}
        self.shutting_down: bool = False
        self.time_last_cleaned_processes: float = time()
        self.apps: Dict[str, ASGIApp] = {"/": app} if app else {}
        for route, app_path in self.config.APPS.items():
            module_str, app_str = app_path.split(":")
            app_module = import_module(module_str)
            self.apps[route] = getattr(app_module, app_str)
        if self.apps:
            self.start_up()
            self.serve_forever()

    @property
    def app(self) -> ASGIApp:
        current_path = self.request_data.path
        for route, app in self.apps.items():
            if route == "/":
                if current_path == route:
                    return app
            elif current_path.startswith(route):
                self.request_data.path = current_path.replace(route, "")
                return app
        app = self.apps.get("/")
        if app:
            return app
        raise HTTPError(404)

    def add_app(self, application: ASGIApp, route: str = "/") -> None:
        self.apps[route] = application

    def serve_forever(self) -> None:
        try:
            while True:
                ready_to_read, _, _ = select.select(
                    [self.listen_socket], [], [], self.config.SELECT_SLEEP_TIME
                )
                if ready_to_read:
                    self.check_socket()
                self.check_processes()
        except KeyboardInterrupt:
            self.shut_down()
        except Exception as err:
            self.logger.exception(err)

    def check_socket(self) -> None:
        try:
            connection, _ = self.listen_socket.accept()
        except IOError as err:
            if err.args[0] != errno.EINTR:
                raise
        else:
            if len(self.processes) > self.config.MAX_PROCESSES:
                self.response.status = b"503"
                self.response.body.write(b"Too many requests")
                connection.settimeout(self.config.RECV_TIMEOUT)
                self.client_connection = connection
                self.finish_response()
                self.response.status = b"200"
                self.response.body.clear()
            elif connection:
                connection.settimeout(self.config.RECV_TIMEOUT)
                self.client_connection = connection
                self.fork()

    def fork(self) -> None:
        process = multiprocessing.Process(target=self.handle_one_request)
        process.daemon = True
        try:
            process.start()
        except AttributeError:
            self.logger.warning(f"Could not start process {process.ident}")
        else:
            self.processes[process.ident] = process

    def check_processes(self) -> None:
        current_time = time()
        if (
            current_time - self.time_last_cleaned_processes
            > self.config.CHECK_PROCESS_INTERVAL
        ):
            self.time_last_cleaned_processes = current_time
            for ident, process in list(self.processes.items()):
                if not process.is_alive():
                    process.close()
                    del self.processes[ident]

    def send_to_all_apps(self, scope: Scope, receive: Receive, send: Send) -> None:
        for app in self.apps.values():
            self.loop.run_until_complete(app(scope, receive, send))

    def start_up(self) -> None:
        self.send_to_all_apps(LIFESPAN_SCOPE, self.lifespan_receive, self.send)
        self.logger.info(
            f"Qactuar: Serving {self.scheme.upper()} on {self.host}:{self.port}"
        )

    def shut_down(self) -> None:
        self.shutting_down = True
        self.logger.info("Shutting Down")
        self.send_to_all_apps(LIFESPAN_SCOPE, self.lifespan_receive, self.send)
        sys.exit(0)

    # TODO: http.disconnect when socket closes

    def handle_one_request(self) -> None:
        try:
            self.get_request_data()
            if not self.raw_request_data:
                self.close_socket()
                return
            if (
                self.request_data.headers["Connection"]
                and self.request_data.headers["Upgrade"]
            ):
                self.ws_shake_hand()
            app = self.app
            self.loop.run_until_complete(
                app(self.create_http_scope(), self.receive, self.send)
            )
            self.finish_response()
        except KeyboardInterrupt:
            sys.exit(0)
        except HTTPError as err:
            self.response.status = str(err.args[0]).encode("utf-8")
            self.response.body.write(str(err.args[0]).encode("utf-8"))
            self.finish_response()
        except Exception as err:
            self.logger.exception(err)
            self.response.status = b"500"
            self.response.body.write(b"wat happen?!")
            self.finish_response()

    def get_request_data(self) -> None:
        request_data = BytesList()
        request = Request()

        while True:
            try:
                request_data.write(self.client_connection.recv(self.config.RECV_BYTES))
            except socket.timeout:
                request.raw_request = request_data.read()
                if request.headers_complete:
                    content_length = request.headers["content-length"]
                    if content_length is not None and request.command != "GET":
                        if len(request.body) == int(content_length):
                            break
                        else:
                            continue
                    break

        self.request_data = request
        self.raw_request_data = request_data.read()

    async def send(self, data: Message) -> None:
        if data["type"] == "http.response.start":
            self.response.status = str(data["status"]).encode("utf-8")
            self.response.headers = data["headers"]
        if data["type"] == "http.response.body":
            # TODO: check "more_body" and if true then do self.client_connection.send()
            #  the current data
            self.response.body.write(data["body"])
        if (
            data["type"] == "lifespan.startup.failed"
            or data["type"] == "lifespan.shutdown.failed"
        ):
            if "startup" in data["type"]:
                self.logger.error("App startup failed")
            if "shutdown" in data["type"]:
                self.logger.error("App shutdown failed")
            self.logger.error(data["message"])

    async def receive(self) -> Message:
        # TODO: support streaming from client
        return {
            "type": "http.request",
            "body": self.request_data.body,
            "more_body": False,
        }

    async def lifespan_receive(self) -> Message:
        return {
            "type": "lifespan.startup"
            if not self.shutting_down
            else "lifespan.shutdown",
            "asgi": ASGI_VERSION,
        }

    def ws_shake_hand(self):
        pass

    def create_websocket(self):
        pass

    def create_http_scope(self) -> Scope:
        # TODO: Pseudo headers (present in HTTP/2 and HTTP/3) must be removed; if
        #  :authority is present its value must be added to the start of the iterable
        #  with host as the header name or replace any existing host header already
        #  present.
        return {
            "type": "http",
            "asgi": ASGI_VERSION,
            "http_version": self.request_data.request_version_num,
            "method": self.request_data.command,
            "scheme": self.scheme,
            "path": self.request_data.path,
            "raw_path": self.request_data.raw_path,
            "query_string": self.request_data.query_string,
            "root_path": "",
            "headers": self.request_data.raw_headers,
            "client": self.client_connection.getpeername(),
            "server": (self.server_name, self.server_port),
        }

    def finish_response(self) -> None:
        try:
            self.client_connection.sendall(self.response.to_http())
        except OSError as err:
            self.logger.exception(err)
        finally:
            self.close_socket()

    def close_socket(self) -> None:
        try:
            self.client_connection.shutdown(socket.SHUT_RDWR)
        except OSError:
            pass
        self.client_connection.close()
